@@ -102,6 +102,7 @@ EXPORT const char nvnc_version[] = "UNKNOWN";
 #endif
 
 extern const unsigned short code_map_qnum_to_linux[];
+extern const unsigned int code_map_qnum_to_linux_len;
 
 static uint64_t nvnc__htonll(uint64_t x)
 {
@@ -163,6 +164,13 @@ static void client_drain_encoder(struct nvnc_client* client)
 
 static void client_close(struct nvnc_client* client)
 {
+	if (client->close_task) {
+		struct aml_idle* task = client->close_task;
+		client->close_task = NULL;
+		aml_stop(aml_get_default(), task);
+		aml_unref(task);
+	}
+
 	nvnc_log(NVNC_LOG_INFO, "Closing client connection %p", client);
 
 	stream_close(client->net_stream);
@@ -202,43 +210,36 @@ static void client_close(struct nvnc_client* client)
 	free(client);
 }
 
-static void do_deferred_client_close(void *obj)
+static void do_deferred_client_close(void* obj)
 {
-	client_close(obj);
-}
+	struct aml_idle* idle = obj;
+	struct nvnc_client* client = aml_get_userdata(idle);
+	client->close_task = NULL;
+	aml_stop(aml_get_default(), idle);
+	aml_unref(idle);
 
-static void stop_self(void* obj)
-{
-	aml_stop(aml_get_default(), obj);
+	client_close(client);
 }
 
 static void defer_client_close(struct nvnc_client* client)
 {
-	struct aml_idle* idle = aml_idle_new(stop_self, client,
-			do_deferred_client_close);
-	aml_start(aml_get_default(), idle);
-	aml_unref(idle);
+	if (client->close_task)
+		return;
+	client->close_task = aml_idle_new(do_deferred_client_close, client,
+			NULL);
+	aml_start(aml_get_default(), client->close_task);
 }
 
 void close_after_write(void* userdata, enum stream_req_status status)
 {
-	struct nvnc_client* client = userdata;
-	nvnc_log(NVNC_LOG_DEBUG, "close_after_write(%p)", client);
-	stream_close(client->net_stream);
-
-	/* This is a rather hacky way of making sure that the client object
-	 * stays alive while the stream is processing its queue.
-	 * TODO: Figure out some better resource management for clients
-	 */
-	defer_client_close(client);
+	struct stream* stream = userdata;
+	stream_destroy(stream);
 }
 
 static int handle_unsupported_version(struct nvnc_client* client)
 {
 	char buffer[256];
 
-	client->state = VNC_CLIENT_STATE_ERROR;
-
 	struct rfb_error_reason* reason = (struct rfb_error_reason*)(buffer + 1);
 
 	static const char reason_string[] = "Unsupported version\n";
@@ -249,9 +250,13 @@ static int handle_unsupported_version(struct nvnc_client* client)
 
 	size_t len = 1 + sizeof(*reason) + strlen(reason_string);
 	stream_write(client->net_stream, buffer, len, close_after_write,
-			client);
+			client->net_stream);
 
-	return 0;
+	// Keep stream alive until the result has been sent to the client
+	stream_ref(client->net_stream);
+
+	client_close(client);
+	return -1;
 }
 
 static void init_security_types(struct nvnc* server)
@@ -417,7 +422,7 @@ static void disconnect_all_other_clients(struct nvnc_client* client)
 
 }
 
-static void send_server_init_message(struct nvnc_client* client)
+static int send_server_init_message(struct nvnc_client* client)
 {
 	struct nvnc* server = client->server;
 	struct nvnc_display* display = server->display;
@@ -471,10 +476,11 @@ static void send_server_init_message(struct nvnc_client* client)
 
 	client->known_width = width;
 	client->known_height = height;
-	return;
+	return 0;
 
 close:
 	nvnc_client_close(client);
+	return -1;
 }
 
 static int on_init_message(struct nvnc_client* client)
@@ -488,7 +494,8 @@ static int on_init_message(struct nvnc_client* client)
 
 	update_min_rtt(client);
 
-	send_server_init_message(client);
+	if (send_server_init_message(client) == -1)
+		return -1;
 
 	nvnc_client_fn fn = client->server->new_client_fn;
 	if (fn)
@@ -1043,7 +1050,14 @@ static int on_client_qemu_key_event(struct nvnc_client* client)
 	int down_flag = msg->down_flag;
 	uint32_t xt_keycode = ntohl(msg->keycode);
 
-	uint32_t evdev_keycode = code_map_qnum_to_linux[xt_keycode];
+
+	uint32_t evdev_keycode = 0;
+	if (xt_keycode < code_map_qnum_to_linux_len) {
+		evdev_keycode = code_map_qnum_to_linux[xt_keycode];
+	} else {
+		nvnc_log(NVNC_LOG_WARNING, "Received too large key code from client: %" PRIu32,
+				xt_keycode);
+	}
 	if (!evdev_keycode)
 		evdev_keycode = xt_keycode;
 
@@ -1070,7 +1084,7 @@ static int on_client_qemu_event(struct nvnc_client* client)
 	nvnc_log(NVNC_LOG_WARNING, "Got uninterpretable qemu message from client: %p",
 			client);
 	nvnc_client_close(client);
-	return 0;
+	return -1;
 }
 
 static int on_client_pointer_event(struct nvnc_client* client)
@@ -1175,7 +1189,7 @@ static char* crlf_to_lf(const char* src, size_t len)
 			continue;
 		}
 
-		if ((in_len == 0) || (*(in + 1) != '\n'))
+		if ((in_len == 1) || (*(in + 1) != '\n'))
 			*out++ = '\n';
 
 		in++;
@@ -1291,12 +1305,12 @@ static int process_client_ext_clipboard(struct nvnc_client* client)
 		nvnc_log(NVNC_LOG_ERROR, "Extended clipboard payload length (%d) is greater than max supported length (%d)",
 				length, max_length);
 		nvnc_client_close(client);
-		return 0;
+		return -1;
 	}
 
 	size_t msg_size = sizeof(*msg) + length;
 
-	/* this is expected to be a provide message. if not, tell
+	/* this is expected to be a text provide message. if not, tell
 	 * process_big_cut_text to ignore it, to avoid unnecessarily attempting
 	 * to inflate garbage */
 	if (msg_size > left_to_process) {
@@ -1305,7 +1319,7 @@ static int process_client_ext_clipboard(struct nvnc_client* client)
 		if (!client->cut_text.buffer) {
 			nvnc_log(NVNC_LOG_ERROR, "OOM: %m");
 			nvnc_client_close(client);
-			return 0;
+			return -1;
 		}
 
 		size_t partial_size = left_to_process - sizeof(*msg);
@@ -1316,7 +1330,8 @@ static int process_client_ext_clipboard(struct nvnc_client* client)
 		client->cut_text.length = length;
 		client->cut_text.index = partial_size;
 
-		client->cut_text.is_provide = (flags & RFB_EXT_CLIPBOARD_ACTION_PROVIDE &&
+		client->cut_text.is_text_provide = (flags & RFB_EXT_CLIPBOARD_ACTION_PROVIDE &&
+				flags & RFB_EXT_CLIPBOARD_FORMAT_TEXT &&
 				!(flags & RFB_EXT_CLIPBOARD_CAPS));
 
 		return left_to_process;
@@ -1366,7 +1381,7 @@ static int process_client_cut_text(struct nvnc_client* client)
 		nvnc_log(NVNC_LOG_ERROR, "Copied text length (%d) is greater than max supported length (%d)",
 				length, max_length);
 		nvnc_client_close(client);
-		return 0;
+		return -1;
 	}
 
 	size_t msg_size = sizeof(*msg) + length;
@@ -1385,7 +1400,7 @@ static int process_client_cut_text(struct nvnc_client* client)
 	if (!client->cut_text.buffer) {
 		nvnc_log(NVNC_LOG_ERROR, "OOM: %m");
 		nvnc_client_close(client);
-		return 0;
+		return -1;
 	}
 
 	size_t partial_size = left_to_process - sizeof(*msg);
@@ -1393,7 +1408,7 @@ static int process_client_cut_text(struct nvnc_client* client)
 	memcpy(client->cut_text.buffer, msg->text, partial_size);
 
 	client->cut_text.is_zlib = false;
-	client->cut_text.is_provide = false;
+	client->cut_text.is_text_provide = false;
 	client->cut_text.length = length;
 	client->cut_text.index = partial_size;
 
@@ -1479,7 +1494,7 @@ static void process_big_cut_text(struct nvnc_client* client)
 		return;
 
 	if (client->cut_text.is_zlib) {
-		if (client->cut_text.is_provide)
+		if (client->cut_text.is_text_provide)
 			process_client_ext_clipboard_provide(client,
 					(unsigned char*)client->cut_text.buffer,
 					client->cut_text.length);
@@ -1522,6 +1537,7 @@ static void ext_clipboard_save_provide_msg(struct nvnc* server, const char* text
 	server->ext_clipboard_provide_msg.buffer = malloc(length);
 	if (!server->ext_clipboard_provide_msg.buffer) {
 		nvnc_log(NVNC_LOG_ERROR, "OOM: %m");
+		free(provide_msg_buf);
 		return;
 	}
 
@@ -1846,7 +1862,7 @@ static int on_client_fence(struct nvnc_client* client)
 		nvnc_log(NVNC_LOG_WARNING,
 				"Client sent too long fence message. Closing.");
 		nvnc_client_close(client);
-		return 0;
+		return -1;
 	}
 
 	enum rfb_fence_flags flags = ntohl(msg->flags);
@@ -1897,14 +1913,15 @@ static int on_client_message(struct nvnc_client* client)
 	nvnc_log(NVNC_LOG_WARNING, "Got uninterpretable message from client: %p",
 			client);
 	nvnc_client_close(client);
-	return 0;
+	return -1;
 }
 
 static int try_read_client_message(struct nvnc_client* client)
 {
+	if (client->net_stream->state == STREAM_STATE_CLOSED)
+		return -1;
+
 	switch (client->state) {
-	case VNC_CLIENT_STATE_ERROR:
-		return client->buffer_len - client->buffer_index;
 	case VNC_CLIENT_STATE_WAITING_FOR_VERSION:
 		return on_version_message(client);
 	case VNC_CLIENT_STATE_WAITING_FOR_SECURITY:
@@ -1981,6 +1998,9 @@ static void on_client_event(struct stream* stream, enum stream_event event)
 		if (rc == 0)
 			break;
 
+		if (rc == -1)
+			return;
+
 		client->buffer_index += rc;
 	}
 
@@ -2097,6 +2117,8 @@ static void sockaddr_to_string(char* dst, size_t sz, const struct sockaddr* addr
 		nvnc_log(NVNC_LOG_DEBUG,
 				"Don't know how to convert sa_family %d to string",
 				addr->sa_family);
+		if (sz > 0)
+			*dst = 0;
 		break;
 	}
 }
@@ -2297,6 +2319,7 @@ void nvnc_close(struct nvnc* self)
 
 	nvnc_fb_release(self->cursor.buffer);
 	nvnc_fb_unref(self->cursor.buffer);
+	self->cursor.buffer = NULL;
 
 	// The stream is closed first to stop all communication and to make sure
 	// that encoding of new frames does not start.
@@ -2431,6 +2454,7 @@ static void finish_fb_update(struct nvnc_client* client,
 		if (!client_supports_resizing(client)) {
 			nvnc_log(NVNC_LOG_ERROR, "Display has been resized but client does not support resizing.  Closing.");
 			client_close(client);
+			return;
 		}
 	}
 
